from typing import List, Set

from fastapi import APIRouter, Depends, HTTPException, Path
from pydantic import ValidationError
from starlette import status
from starlette.responses import Response

from id_broker_common.kelvin import kelvin_session
from provisioning_plugin.dependencies import check_for_authority_admin, get_school_class_by_id
from provisioning_plugin.models import NonEmptyStr, NoStarStr, SchoolClass
from provisioning_plugin.utils import get_source_uid_for_authority, search_kelvin_obj
from ucsschool.kelvin.client import (
    InvalidRequest,
    NoObject,
    SchoolClass as KelvinSchoolClass,
    Session,
    UserResource,
)

router = APIRouter(tags=["school_classes"], dependencies=[Depends(check_for_authority_admin)])


async def member_ids_to_usernames(
    members: Set[str], school_authority: str, session: Session
) -> List[str]:
    """
    Takes a list of user_ids and returns a list of corresponding usernames.

    :param members: The list of user_ids to fetch the usernames for
    :param school_authority: The school authority to search in
    :param session: The Kelvin session to search with
    :return: The list of usernames
    :raises NoObject: If one of the members cannot be found
    """
    return [
        (
            await search_kelvin_obj(
                UserResource(session=session),
                record_uid=user_id,
                source_uid=get_source_uid_for_authority(school_authority),
            )
        ).name
        for user_id in members
    ]


@router.get("/classes/{id}", response_model=SchoolClass)
async def get(
    class_id: NoStarStr = Path(
        ...,
        alias="id",
        description="Unique ID of LDAP object on school authority side.",
        title="Object ID",
    ),
    school_authority: NonEmptyStr = Path(
        ...,
        description="Identifier of the school authority this object originates from.",
        title="School authority ID",
    ),
    school_class: KelvinSchoolClass = Depends(get_school_class_by_id),
    session: Session = Depends(kelvin_session),
) -> SchoolClass:
    try:
        return SchoolClass.without_prefix(
            await SchoolClass.from_kelvin_school_class(
                school_class.as_dict(), UserResource(session=session)
            ),
            school_authority,
        )
    except (ValidationError, ValueError, NoObject) as exc:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"The school class {class_id!r} in school authority {school_authority!r} is "
            f"malformed.",
        ) from exc


@router.head("/classes/{id}")
async def get_head(
    class_id: NoStarStr = Path(
        ...,
        alias="id",
        description="Unique ID of LDAP object on school authority side.",
        title="Object ID",
    ),
    school_authority: NonEmptyStr = Path(
        ...,
        description="Identifier of the school authority this object originates from.",
        title="School authority ID",
    ),
    school_class: KelvinSchoolClass = Depends(get_school_class_by_id),
    session: Session = Depends(kelvin_session),
):
    await get(class_id, school_authority, school_class, session)  # will raise 404 if it doesn't exist
    return Response(status_code=status.HTTP_200_OK)


@router.post("/classes", response_model=SchoolClass, status_code=201)
async def post(
    school_class_data: SchoolClass,
    school_authority: NonEmptyStr = Path(
        ...,
        description="Identifier of the school authority this object originates from.",
        title="School authority ID",
    ),
    session: Session = Depends(kelvin_session),
) -> SchoolClass:
    school_class_data = SchoolClass.with_prefix(school_class_data, school_authority)
    try:
        usernames = await member_ids_to_usernames(school_class_data.members, school_authority, session)
    except NoObject as exc:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"The school class {school_class_data.name!r} in school {school_class_data.school!r} "
            f"in school authority {school_authority!r} is malformed.",
        ) from exc
    school_class = KelvinSchoolClass(
        name=school_class_data.name,
        school=school_class_data.school,
        users=usernames,
        description=school_class_data.description,
        udm_properties={
            "ucsschoolRecordUID": school_class_data.id,
            "ucsschoolSourceUID": school_authority,
        },
        session=session,
    )
    try:
        await school_class.save()
    except InvalidRequest as exc:
        raise HTTPException(status_code=exc.status, detail=exc.reason) from exc
    # To avoid retrieving all members again, at this point we trust that the data was saved correctly
    # TODO: If performance problem is solved, use created user, not input
    return SchoolClass.without_prefix(school_class_data, school_authority)


@router.put("/classes/{id}", response_model=SchoolClass)
async def put(
    school_class_data: SchoolClass,
    class_id: NoStarStr = Path(
        ...,
        alias="id",
        description="Unique ID of LDAP object on school authority side.",
        title="Object ID",
    ),
    school_authority: NonEmptyStr = Path(
        ...,
        description="Identifier of the school authority this object originates from.",
        title="School authority ID",
    ),
    school_class: KelvinSchoolClass = Depends(get_school_class_by_id),
    session: Session = Depends(kelvin_session),
):
    if school_class_data.school != school_class.school:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="The school must not be changed after creation.",
        )
    school_class_data = SchoolClass.with_prefix(school_class_data, school_authority)
    school_class.name = school_class_data.name
    school_class.description = school_class_data.description
    try:
        usernames = await member_ids_to_usernames(school_class_data.members, school_authority, session)
    except NoObject as exc:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail="Some of the users in the members field do not exist.",
        ) from exc
    school_class.users = usernames
    try:
        await school_class.save()
    except InvalidRequest as exc:
        raise HTTPException(status_code=exc.status, detail=exc.reason) from exc
    return SchoolClass.without_prefix(school_class_data, school_authority)


@router.delete("/classes/{id}", status_code=status.HTTP_204_NO_CONTENT)
async def delete(
    class_id: NoStarStr = Path(
        ...,
        alias="id",
        description="Unique ID of LDAP object on school authority side.",
        title="Object ID",
    ),
    school_authority: NonEmptyStr = Path(
        ...,
        description="Identifier of the school authority this object originates from.",
        title="School authority ID",
    ),
    school_class: KelvinSchoolClass = Depends(get_school_class_by_id),
) -> Response:
    await school_class.delete()
    return Response(status_code=status.HTTP_204_NO_CONTENT)
